@@ -1,6 +1,6 @@
 package Dancer2::Plugin::Auth::Extensible;
 
-our $VERSION = '0.620';
+our $VERSION = '0.621';
 
 use strict;
 use warnings;
@@ -241,16 +241,18 @@ sub BUILD {
 
                 if ( $weak_plugin->logged_in_user ) {
                     # User is already logged in so redirect elsewhere
+                    # uncoverable condition false
                     $app->redirect(
-                             $app->request->query_parameter->get('return_url')
+                             $app->request->query_parameters->get('return_url')
                           || $weak_plugin->user_home_page );
                 }
 
                 # Reset password code submitted?
                 my ($code) = $app->request->splat;
 
-                if (   $weak_plugin->reset_password_handler
-                    && $weak_plugin->user_password(code => $code ) )
+                if (   $code
+                    && $weak_plugin->reset_password_handler
+                    && $weak_plugin->user_password( code => $code ) )
                 {
                     $app->request->params->{password_code_valid} = 1;
                 }
@@ -303,8 +305,13 @@ sub auth_provider {
     # If no realm was provided, but we have a logged in user, use their realm.
     # Don't try and read the session any earlier though, as it won't be
     # available on plugin load
-    if ( !$realm && $plugin->app->session->read('logged_in_user') ) {
-        $realm = $plugin->app->session->read('logged_in_user_realm');
+    if ( !defined $realm ) {
+        if ( $plugin->app->session->read('logged_in_user') ) {
+            $realm = $plugin->app->session->read('logged_in_user_realm');
+        }
+        else {
+            croak "auth_provider needs realm or there must be a logged in user";
+        }
     }
 
     # First, if we already have a provider for this realm, go ahead and use it:
@@ -364,6 +371,7 @@ sub authenticate_user {
                     %lastlogin );
                 1;
             } or do {
+                # uncoverable condition right
                 my $err = $@ || "Unknown error";
                 $plugin->app->log(
                     error => "$realm provider threw error: $err" );
@@ -404,12 +412,14 @@ sub create_user {
 
     $plugin->execute_plugin_hook( 'before_create_user', \%options );
 
+    # uncoverable condition false
     my $realm         = delete $options{realm} || $plugin->realm_names->[0];
     my $email_welcome = delete $options{email_welcome};
     my $password      = delete $options{password};
     my $provider      = $plugin->auth_provider($realm);
 
     eval { $user = $provider->create_user(%options); 1; } or do {
+        # uncoverable condition right
         my $err = $@ || "Unknown error";
         $plugin->app->log( error => "$realm provider threw error: $err" );
         push @errors, $err;
@@ -437,6 +447,7 @@ sub create_user {
                 $provider->set_user_password( $options{username}, $password );
                 1;
             } or do {
+                # uncoverable condition right
                 my $err = $@ || "Unknown error";
                 $plugin->app->log(
                     error => "$realm provider threw error: $err" );
@@ -463,6 +474,7 @@ sub get_user_details {
             debug => "Attempting to find user $username in realm $realm" );
         my $provider = $plugin->auth_provider($realm);
         eval { $user = $provider->get_user_details($username); 1; } or do {
+            # uncoverable condition right
             my $err = $@ || "Unknown error";
             $plugin->app->log( error => "$realm provider threw error: $err" );
         };
@@ -495,7 +507,16 @@ sub logged_in_user {
 }
 
 sub logged_in_user_lastlogin {
-    shift->app->session->read('logged_in_user_lastlogin');
+    my $lastlogin = shift->app->session->read('logged_in_user_lastlogin');
+    if ( defined $lastlogin && ref($lastlogin) eq '' && $lastlogin =~ /^\d+$/ )
+    {
+        # A sane epoch time. Old Provider::DBIC stores DateTime in the session
+        # which might get stringified or perhaps not and some session engines
+        # might fail to serialize/deserialize so we now store epoch and
+        # convert back to DateTime.
+        $lastlogin = DateTime->from_epoch( epoch => $lastlogin );
+    }
+    return $lastlogin;
 }
 
 sub logged_in_user_password_expired {
@@ -618,7 +639,8 @@ sub update_user {
     croak "Realm must be specified when more than one realm configured"
       if !$update{realm} && $plugin->realm_count > 1;
 
-    my $realm    = delete $update{realm} || $plugin->realm_name->[0];
+    # uncoverable condition false
+    my $realm    = delete $update{realm} || $plugin->realm_names->[0];
     my $provider = $plugin->auth_provider($realm);
     my $updated  = $provider->set_user_details( $username, %update );
     my $cur_user = $plugin->app->session->read('logged_in_user');
@@ -737,16 +759,19 @@ sub user_password {
 
 sub user_roles {
     my ( $plugin, $username, $realm ) = @_;
-    croak "Cannot call user_roles since roles are disabled by disable_roles setting"
+    croak
+      "Cannot call user_roles since roles are disabled by disable_roles setting"
       if $plugin->disable_roles;
 
-    $username = $plugin->app->session->read('logged_in_user')
-      unless defined $username;
+    if ( !defined $username ) {
+        # assume logged_in_user so clear realm and look for user
+        $realm = undef;
+        $username = $plugin->app->session->read('logged_in_user');
+        croak "user_roles needs a username or a logged in user"
+          unless $username;
+    }
 
-    my $search_realm = ( $realm ? $realm : '' );
-
-    my $roles =
-      $plugin->auth_provider($search_realm)->get_user_roles($username);
+    my $roles = $plugin->auth_provider($realm)->get_user_roles($username);
     return unless defined $roles;
     return wantarray ? @$roles : $roles;
 }
@@ -1014,28 +1039,6 @@ sub _send_email {
     }
 }
 
-# Given a class method name and a set of parameters, try calling that class
-# method for each realm in turn, arranging for each to receive the configuration
-# defined for that realm, until one returns a non-undef, then return the realm
-# which succeeded and the response.
-# Note: all provider class methods return a single value; if any need to return
-# a list in future, this will need changing)
-#
-# FIXME: this is not used anywhere - what is it for?
-sub _try_realms {
-    my ( $plugin, $method, @args );
-    for my $realm ( @{ $plugin->realm_names } ) {
-        my $provider = $plugin->auth_provider($realm);
-        if ( !$provider->can($method) ) {
-            croak "Provider $provider does not provide a $method method!";
-        }
-        if ( defined( my $result = $provider->$method(@args) ) ) {
-            return $result;
-        }
-    }
-    return;
-}
-
 #
 # routes
 #
@@ -1120,6 +1123,7 @@ sub _post_login_route {
     }
 
     if ( $plugin->logged_in_user ) {
+        # uncoverable condition false
         $app->redirect( $app->request->params->{return_url}
               || $plugin->user_home_page );
     }
@@ -1138,6 +1142,7 @@ sub _post_login_route {
         $app->session->write( logged_in_user_realm => $realm );
         $app->log( core => "Realm is $realm" );
         $plugin->execute_plugin_hook( 'after_login_success' );
+        # uncoverable condition false
         $app->redirect( $app->request->params->{return_url}
               || $plugin->user_home_page );
     }
@@ -1208,7 +1213,7 @@ Configure the plugin to use the authentication provider class you wish to use:
         Auth::Extensible:
             realms:
                 users:
-                    provider: Example
+                    provider: Config
                     ....
 
 The configuration you provide will depend on the authentication provider module
@@ -1236,7 +1241,7 @@ to authenticate a user against the chosen source of authentication.
 
 For an example of how simple provider classes are, so you can build your own if
 required or just try out this authentication framework plugin easily, 
-see L<Dancer2::Plugin::Auth::Extensible::Provider::Example>.
+see L<Dancer2::Plugin::Auth::Extensible::Provider::Config>.
 
 This framework supplies the following providers out-of-the-box:
 
