@@ -1,6 +1,6 @@
 package Dancer2::Plugin::Auth::Extensible::Test;
 
-our $VERSION = '0.620';
+our $VERSION = '0.621';
 
 =head1 NAME
 
@@ -14,6 +14,7 @@ use strict;
 use Carp qw(croak);
 use Test::More;
 use Test::Deep;
+use Test::MockDateTime;
 use Plack::Test;
 use HTTP::Cookies;
 use HTTP::Request::Common qw(GET POST);
@@ -50,6 +51,7 @@ my %dispatch = (
     logged_in_user                  => \&_logged_in_user,
     logged_in_user_lastlogin        => \&_logged_in_user_lastlogin,
     logged_in_user_password_expired => \&_logged_in_user_password_expired,
+    password_reset                  => \&_password_reset,
     password_reset_send             => \&_password_reset_send,
     require_login                   => \&_require_login,
     roles                           => \&_roles,
@@ -62,15 +64,16 @@ my %dispatch = (
 # These are assumed to be correct. If they are not then some provider tests
 # should fail and we can fixup later.
 my %dependencies = (
-    authenticate_user => ['authenticate_user'],
     create_user => [ 'get_user_details', 'create_user', 'set_user_details', ],
     get_user_details => ['get_user_details'],
     logged_in_user   => ['get_user_details'],
+    logged_in_user_lastlogin => ['create_user','record_lastlogin'],
     logged_in_user_password_expired =>
       [ 'get_user_details', 'password_expired' ],
+    password_reset      => ['get_user_by_code'],
     password_reset_send => ['set_user_details'],
     require_login       => ['get_user_details'],
-    roles               => [ 'get_user_details', 'get_user_roles' ],
+    roles               => ['get_user_roles' ],
     update_current_user => ['set_user_details'],
     update_user         => ['set_user_details'],
     user_password =>
@@ -83,6 +86,9 @@ sub testme {
     BAIL_OUT "Please upgrade your provider to the latest version. Dancer2::Plugin::Auth::Extensible no longer supports the old \"testme\" tests.";
 }
 
+# so test can check
+my @provider_can;
+
 sub runtests {
     my $app = shift;
 
@@ -98,7 +104,7 @@ sub runtests {
     BAIL_OUT "Unexpected response to /provider_can"
       unless ref($ret) eq 'ARRAY';
 
-    my @provider_can = @$ret;
+    @provider_can = @$ret;
 
     my @to_test = ($ENV{D2PAE_TEST_ONLY}) || keys %dispatch;
 
@@ -111,6 +117,10 @@ sub runtests {
             skip "Provider $test tests as provider is missing methods: "
               . join( ", ", @missing ), 1
               if @missing;
+
+            # for safety in case one set of tests doesn't clean up carefully
+            $jar->clear;
+
             subtest "Plugin $test tests" => $dispatch{$test};
         }
     }
@@ -144,10 +154,9 @@ sub post {
 sub _authenticate_user {
     my ($res, $data, $logs);
 
-    $trap->read;
-
     # no args
 
+    $trap->read;
     $res = post('/authenticate_user');
     ok $res->is_success, "/authenticate_user with no params is_success";
     cmp_deeply YAML::Load( $res->content ), [ 0, undef ],
@@ -327,7 +336,6 @@ sub _authenticate_user {
       "... and the other realms were not checked."
       or diag explain $logs;
 
-
     # good username, good password and no realm
 
     $res = post( '/authenticate_user',
@@ -425,6 +433,17 @@ sub _authenticate_user {
       ),
       "... and we don't see realm config1 checked."
       or diag explain $logs;
+
+    # quick pairwise for coverage
+    foreach my $username ( undef, +{}, '', 'username' ) {
+        foreach my $password ( undef, +{}, '', 'password' ) {
+            $res = post( '/authenticate_user',
+                [ username => $username, password => $password ] );
+            ok $res->is_success, "/authenticate_user with user dave, bad password and no realm success";
+            cmp_deeply YAML::Load( $res->content ), [ 0, undef ],
+              "... and returns expected response";
+        }
+    }
 }
 
 #------------------------------------------------------------------------------
@@ -457,6 +476,12 @@ sub _create_user {
       ],
       "... and error about needing realm was logged.";
 
+    # create user with no password
+
+    $res = post( "/create_user",
+        [ username => 'createusernopw', realm => 'config1' ] );
+    ok $res->is_success, "/create_user with no password is_success";
+
     for my $realm (qw/config1 config2/) {
 
         # create a user
@@ -665,10 +690,44 @@ sub _get_user_details {
 #
 #  login_logout
 #
+#  also includes some auth_provider tests
+#
 #------------------------------------------------------------------------------
 
 sub _login_logout {
-    my ( $data, $res );
+    my ( $data, $res, $logs );
+
+    # auth_provider with no args
+
+    $trap->read;
+    $res = post('/auth_provider');
+    is $res->code, 500, "auth_provider with no args dies";
+    $logs = $trap->read;
+    cmp_deeply $logs, superbagof(
+        { formatted => ignore(),
+            level => 'error',
+            message => re(qr/auth_provider needs realm or/),
+        }
+    ), "... and correct error message is seen in logs." or diag explain $logs;
+
+    # auth_provider with non-existant realm
+
+    $trap->read;
+    $res = post('/auth_provider', [realm => 'NoSuchRealm']);
+    is $res->code, 500, "auth_provider with non-existant realm dies";
+    $logs = $trap->read;
+    cmp_deeply $logs, superbagof(
+        { formatted => ignore(),
+            level => 'error',
+            message => re(qr/Invalid realm NoSuchRealm/),
+        }
+    ), "... and correct error message is seen in logs." or diag explain $logs;
+
+    # auth_provider with good realm
+
+    $res = post('/auth_provider', [realm => 'config1']);
+    ok $res->is_success, "auth_provider with good realm lives"
+      or diag explain $trap->read;
 
     # Check that login route doesn't match any request string with '/login'.
 
@@ -704,6 +763,9 @@ sub _login_logout {
     }
     like $res->content, qr/input.+name="password"/,
       "... and we have the login page.";
+    like $res->content,
+      qr/Enter your username to obtain an email to reset your password/,
+      "... which has password reset option (reset_password_handler=>1).";
 
     # post empty /login
 
@@ -744,6 +806,28 @@ sub _login_logout {
     is $data->{logged_in_user_realm}, 'config1',
       "... and session logged_in_user_realm is set to config1.";
 
+    # get /login whilst already logged in
+
+    $res = get('/login');
+    ok $res->is_redirect, "GET /login whilst logged in is redirected."
+      or diag explain $res;
+    is $res->header('location'), 'http://localhost/',
+      "... and redirect location is correct.";
+
+    # get /login whilst already logged in with return_url set
+
+    $res = get('/login?return_url=/foo');
+    ok $res->is_redirect,
+      "GET /login whilst logged in with return_url set in query is redirected.";
+    is $res->header('location'), 'http://localhost/foo',
+      "... and redirect location is correct.";
+
+    # auth_provider with no realm but user is logged in
+
+    $res = post('/auth_provider');
+    ok $res->is_success, "auth_provider with *no* realm lives"
+      or diag explain $trap->read;
+
     # get /logout
 
     $res = get('/logout');
@@ -778,12 +862,13 @@ sub _login_logout {
     is $data->{logged_in_user_realm}, 'config1',
       "... and session logged_in_user_realm is set to config1.";
 
-    # POST /logout
+    # POST /logout with return_url
 
-    $res = post('/logout');
-    ok $res->is_redirect, "POST /logout is_redirect" or diag explain $res;
-    is $res->header('location'), 'http://localhost/',
-      "... and redirect location is correct.";
+    $res = post('/logout', [ return_url => '/foo/bar' ] );
+    ok $res->is_redirect, "POST /logout with return_url /foo/bar is_redirect"
+      or diag explain $res;
+    is $res->header('location'), 'http://localhost/foo/bar',
+      "... and redirect location /foo/bar is correct.";
 
     # check session_data again
 
@@ -889,6 +974,15 @@ sub _login_logout {
             'http://localhost/',
             '/logout redirected to / (exit_page) after logging out' );
     }
+
+    # /login/denied page
+
+    {
+        my $res = get('/login/denied');
+        is $res->code, '403', "GET /login/denied results in a 403 denied code";
+        like $res->content, qr/Permission Denied/,
+          "... and we have Permission Denied text in page";
+    }
 }
 
 #------------------------------------------------------------------------------
@@ -943,6 +1037,14 @@ sub _logged_in_user {
     is $data->{name}, 'David Precious',
       "... and we see dave's name is David Precious." or diag explain $data;
 
+    # check logged_in_user gets cached (coverage)
+
+    $res = get('/logged_in_user_twice');
+    ok $res->is_success, "/logged_in_user_twice is_success";
+    $data = YAML::Load $res->content;
+    is $data->{name}, 'David Precious',
+      "... and we see dave's name is David Precious." or diag explain $data;
+
     # get /logout
 
     $res = get('/logout');
@@ -966,7 +1068,106 @@ sub _logged_in_user {
 #------------------------------------------------------------------------------
 
 sub _logged_in_user_lastlogin {
-    ok 1;
+    my ( $res, $session );
+
+    # create a new user for test so we are sure lastlogin has not been set
+
+    $res = post(
+        "/create_user",
+        [
+            username      => 'lastlogin1',
+            password      => 'lastlogin2',
+            realm         => 'config1',
+        ]
+    );
+    ok $res->is_success, "create_user lastlogin1 call is_success";
+
+    # check the session for logged_in_user_lastlogin
+
+    $res = get('/session_data');
+    ok $res->is_success, "get /session_data is_success";
+    $session = YAML::Load $res->content;
+    ok !defined $session->{logged_in_user_lastlogin},
+      "... and logged_in_user_lastlogin is not set in the session.";
+
+    # we cannot reach require_login routes
+
+    $res = get('/loggedin');
+    ok $res->is_redirect, "GET /loggedin causes redirect";
+    is $res->header('location'),
+      'http://localhost/login?return_url=%2Floggedin',
+      "... and we're redirected to /login with return_url=/loggedin.";
+
+    # login
+
+    $res =
+      post( '/login', [ username => 'lastlogin1', password => 'lastlogin2' ] );
+    ok $res->is_redirect, "POST /login with with new user is_redirect";
+    is $res->header('location'), 'http://localhost/',
+      "... and redirect location is correct.";
+
+    # check we can reach restricted page
+
+    $res = get('/loggedin');
+    ok $res->is_success, "GET /loggedin is_success now we're logged in";
+    is $res->content, "You are logged in", "... and we can see page content.";
+
+    # check the session for logged_in_user_lastlogin
+
+    $res = get('/session_data');
+    ok $res->is_success, "get /session_data is_success";
+    $session = YAML::Load $res->content;
+    ok !defined $session->{logged_in_user_lastlogin},
+      "... and logged_in_user_lastlogin is still not set in the session.";
+
+    # check logged_in_user_lastlogin method
+
+    $res = get('/logged_in_user_lastlogin');
+    ok $res->is_success, "get /logged_in_user_lastlogin is_success";
+    is $res->content, "not set",
+      "... and logged_in_user_lastlogin returns undef"
+      or diag explain $res->content;
+
+    # logout
+
+    $res = get('/logout');
+    ok $res->is_redirect, "/logout is_success";
+
+    # login again and now logged_in_user_lastlogin should be set
+
+    $res =
+      post( '/login', [ username => 'lastlogin1', password => 'lastlogin2' ] );
+    ok $res->is_redirect, "POST /login with with new user is_redirect";
+    is $res->header('location'), 'http://localhost/',
+      "... and redirect location is correct.";
+
+    # check we can reach restricted page
+
+    $res = get('/loggedin');
+    ok $res->is_success, "GET /loggedin is_success now we're logged in";
+    is $res->content, "You are logged in", "... and we can see page content.";
+
+    # check the session for logged_in_user_lastlogin
+
+    $res = get('/session_data');
+    ok $res->is_success, "get /session_data is_success";
+    $session = YAML::Load $res->content;
+    ok defined $session->{logged_in_user_lastlogin},
+      "... and logged_in_user_lastlogin is still not set in the session."
+      or diag explain $session;
+    like $session->{logged_in_user_lastlogin}, qr/^\d+$/,
+      "... and session logged_in_user_lastlogin looks like an epoch time.";
+
+    # check logged_in_user_lastlogin method
+
+    $res = get('/logged_in_user_lastlogin');
+    ok $res->is_success, "get /logged_in_user_lastlogin is_success";
+    my $date = DateTime->now->ymd;
+    is $res->content, $date,
+      "... and logged_in_user_lastlogin is $date.";
+
+    # cleanup
+    get('/logout');
 }
 
 #------------------------------------------------------------------------------
@@ -976,6 +1177,55 @@ sub _logged_in_user_lastlogin {
 #------------------------------------------------------------------------------
 
 sub _logged_in_user_password_expired {
+    my $res;
+
+    my $data = [
+        username => 'pwdexpired1',
+        password => 'pwd1',
+        realm    => 'config1'
+    ];
+
+    on '2016-10-01 00:00:00' => sub {
+        $res = post( '/create_user',$data);
+        ok $res->is_success, "create user pwdexpired1 is success on 2016-10-01";
+        is $res->content, 1,
+          "... and it seems user was created based on response.";
+
+        $res = get('/logged_in_user_password_expired');
+        is $res->content, 'no',
+          "... and before login logged_in_user_password_expired is false.";
+
+        post('/login', $data);
+        $res = get('/loggedin');
+        ok $res->is_success, "User is now logged in";
+
+        $res = get('/logged_in_user_password_expired');
+        is $res->content, 'no',
+          "... and logged_in_user_password_expired is false.";
+    };
+
+    note "... time passes ...";
+
+    on '2016-11-01 00:00:00' => sub {
+
+        $res = get('/logged_in_user_password_expired');
+        is $res->content, 'yes',
+          "... and 30 days later logged_in_user_password_expired is true.";
+
+        post('/logout');
+    };
+}
+
+#------------------------------------------------------------------------------
+#
+#  password_reset
+#
+#------------------------------------------------------------------------------
+
+sub _password_reset {
+    my $res;
+
+    $res = get('/login');
     ok 1;
 }
 
@@ -1087,6 +1337,25 @@ sub _roles {
         );
     }
 
+    {
+        $trap->read;
+        my $res = get('/user_roles');
+        is $res->code, 500,
+          "user_roles with no logged_in_user and no args dies";
+        my $logs = $trap->read;
+        cmp_deeply $logs,
+          superbagof(
+            {
+                formatted => ignore(),
+                level     => 'error',
+                message =>
+                  re(qr/user_roles needs a username or a logged in user/),
+            }
+          ),
+          "got error: user_roles needs a username or a logged in user",
+          or diag explain $logs;
+    }
+
     # and can't reach pages that have require_role
 
     {
@@ -1370,6 +1639,24 @@ sub _roles {
 #------------------------------------------------------------------------------
 
 sub _update_current_user {
+
+    # no user logged in
+
+    $trap->read;
+    my $res = get("/update_current_user");
+    ok $res->is_success, "get /update_current_user is_success"
+      or diag explain $trap->read;
+    cmp_deeply $trap->read,
+      superbagof(
+        {
+            formatted => ignore(),
+            level     => 'debug',
+            message =>
+              'Could not update current user as no user currently logged in',
+        }
+      ),
+      "Could not update current user as no user currently logged in";
+
     for my $realm (qw/config1 config2/) {
 
         # Now we're going to update the current user
@@ -1437,6 +1724,26 @@ sub _update_current_user {
 #------------------------------------------------------------------------------
 
 sub _update_user {
+
+    # update_user with no realm specified
+
+    $trap->read;
+    my $res = post("/update_user", [ username => "mark", name => "FooBar" ]);
+    is $res->code, 500, "update_user with no realm specified croaks 500";
+    my $logs = $trap->read;
+    cmp_deeply $logs,
+      superbagof(
+        {
+            formatted => ignore(),
+            level     => "error",
+            message   => re(
+                qr/Realm must be specified when more than one realm configured/
+            ),
+        }
+      ),
+      "got log: Realm must be specified when more than one realm configured."
+      or diag explain $logs;
+
     for my $realm (qw/config1 config2/) {
 
         # First test a standard user details update.
@@ -1480,6 +1787,48 @@ sub _update_user {
                 "Name is now Wiltshire Apples $realm"
             );
         }
+
+        # log in user dave and whilst logged in change user mark
+        {
+            my $res = post('/login', [username => 'dave', password => 'beer']);
+            ok $res->is_redirect, "login user dave";
+
+            $res = get('/logged_in_user');
+            ok $res->is_success, "... and get logged_in_user is_success";
+            like $res->content, qr/David Precious/,
+              "... and we see dave's details."
+              or diag $res->content;
+
+            # Update mark
+            $res = post(
+                "/update_user",
+                [
+                    realm    => $realm,
+                    username => "mark",
+                    name     => "No beer for me"
+                ]
+            );
+            ok $res->is_success, "change mark's name is success"
+              or diag explain $trap->read;
+
+            $trap->read;    # clear logs
+
+            # check it
+            $res = get("/get_user_mark/$realm");
+            ok $res->is_success, "get /get_user_mark/$realm is_success"
+              or diag explain $trap->read;
+
+            my $user = YAML::Load $res->content;
+            is $user->{name}, "No beer for me",
+              "... and mark's name is now No beer for me.";
+
+            $res = get('/logged_in_user');
+            ok $res->is_success, "get logged_in_user is_success";
+            like $res->content, qr/David Precious/,
+              "... and we see still see dave's details."
+              or diag $res->content;
+
+        }
     }
 }
 
