@@ -6,7 +6,7 @@ Dancer2::Plugin::Auth::Extensible::Test::App - Dancer2 app for testing providers
 
 =cut
 
-our $VERSION = '0.620';
+our $VERSION = '0.621';
 
 use strict;
 use warnings;
@@ -41,6 +41,8 @@ my $provider = $plugin->auth_provider('config1');
 
 my @provider_can = ();
 
+push @provider_can, 'record_lastlogin' if $plugin->config->{record_lastlogin};
+
 #
 # IMPORTANT NOTE
 #
@@ -418,6 +420,36 @@ SKIP: {
     };
 }
 
+subtest "Plugin coverage testing" => sub {
+    # DO NOT use this for testing things that can be tested elsewhere since
+    # these tests are purely to catch the code paths that we can't get to
+    # any other way.
+
+    like exception { $plugin->realm() }, qr/realm name not provided/,
+      "Calling realm method with no args dies";
+
+    like exception { $plugin->realm('') }, qr/realm name not provided/,
+      "... and calling it with single empty arg dies.";
+
+    foreach my $username ( undef, +{}, '', 'username' ) {
+        foreach my $password ( undef, +{}, '', 'password' ) {
+            my $ret = $plugin->authenticate_user( $username, $password );
+            is $ret, 0,
+                "Checking authenticate_user with username/password: "
+              . mydumper($username) . "/"
+              . mydumper($password);
+        }
+    }
+};
+
+sub mydumper {
+    my $val = shift;
+    !defined $val && return '(undef)';
+    ref($val) ne '' && return ref($val);
+    $val eq '' && return '(empty)';
+    $val;
+};
+
 # hooks
 
 hook before_authenticate_user => sub {
@@ -478,6 +510,14 @@ get '/session_data' => sub {
     send_as YAML => $session;
 };
 
+get '/logged_in_user_lastlogin' => sub {
+    my $dt = logged_in_user_lastlogin;
+    if ( ref($dt) eq 'DateTime' ) {
+        return $dt->ymd;
+    }
+    return 'not set';
+};
+
 get '/logged_in_user' => sub {
     my $user = logged_in_user;
     if ( blessed($user) ) {
@@ -492,6 +532,21 @@ get '/logged_in_user' => sub {
     send_as YAML => $user ? $user : 'none';
 };
 
+get '/logged_in_user_twice' => sub {
+    logged_in_user; # retrieve
+    my $user = logged_in_user; # should now be stashed in var
+    if ( blessed($user) ) {
+        if ( $user->isa('DBIx::Class::Row')) {
+            $user = +{ $user->get_columns };
+        }
+        else {
+            # assume some kind of hash-backed object
+            $user = \%$user;
+        }
+    }
+    send_as YAML => $user ? $user : 'none';
+};
+
 get '/loggedin' => require_login sub  {
     "You are logged in";
 };
@@ -520,6 +575,9 @@ get '/roles/:user/:realm' => require_login sub {
     return join ',', sort @{ user_roles($user, $realm) };
 };
 
+get '/user_roles' => sub {
+    return join ',', sort @{ user_roles() };
+};
 
 get '/beer' => require_role BeerDrinker => sub {
     "You can have a beer";
@@ -580,6 +638,12 @@ get '/update_user_name/:realm' => sub {
     YAML::Dump update_user 'mark', realm => $realm, name => "Wiltshire Apples $realm";
 };
 
+post '/update_user' => sub {
+    my $params = body_parameters->as_hashref;
+    my $username = delete $params->{username};
+    send_as YAML => update_user $username, %$params;
+};
+
 get '/get_user_mark/:realm' => sub {
     my $realm = param 'realm';
     content_type 'text/x-yaml';
@@ -596,4 +660,14 @@ get '/get_user_mark/:realm' => sub {
     YAML::Dump $user;
 };
 
+post '/auth_provider' => sub {
+    $plugin->auth_provider( body_parameters->get('realm') );
+    return;
+};
+
+get '/logged_in_user_password_expired' => sub {
+    my $ret = logged_in_user_password_expired;
+    return $ret ? 'yes' : 'no';
+};
+
 1;
